/*******************************************************************************
 * Copyright (C) 2009 Ingo Wassink of University of Twente, Netherlands and
 * The University of Manchester   
 * 
 *  Modifications to the initial code base are copyright of their
 *  respective authors, or their employers as appropriate.
 * 
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public License
 *  as published by the Free Software Foundation; either version 2.1 of
 *  the License, or (at your option) any later version.
 *    
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *    
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 ******************************************************************************/

/**
 * @author Ingo Wassink
 * @author Ian Dunlop
 * @author Alan R Williams
 */
package net.sf.taverna.t2.activities.rshell.views;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.util.List;

import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;

import net.sf.taverna.t2.activities.rshell.RShellPortSymanticTypeBean;
import net.sf.taverna.t2.activities.rshell.RshellActivity;
import net.sf.taverna.t2.activities.rshell.RshellPortTypes;
import net.sf.taverna.t2.activities.rshell.RshellPortTypes.SemanticTypes;
import net.sf.taverna.t2.lang.ui.DialogTextArea;
import net.sf.taverna.t2.workflowmodel.processor.activity.config.ActivityInputPortDefinitionBean;

/**
 * A view representing {@link ActivityInputPortDefinitionBean}s of a
 * {@link RshellActivity} and the various parts which can be edited,
 * primarily the name and the {@link SemanticTypes}.
 * 
 * @author Ian Dunlop
 * 
 */
@SuppressWarnings("serial")
public class RshellInputViewer extends JPanel {

	private ActivityInputPortDefinitionBean bean;

	private JTextField nameField;

	private JSpinner depthSpinner;

	private DialogTextArea refSchemeText;

	private DialogTextArea mimeTypeText;

	private JLabel translatedType;

	private JComboBox semanticTypeSelector;

	private boolean editable;

	private final List<RShellPortSymanticTypeBean> list;

	/**
	 * Calls {@link #initView()} to set the look and feel and sets the
	 * components to be editable or not
	 * 
	 * @param bean
	 *            the {@link ActivityInputPortDefinitionBean} which represents
	 *            the view
	 * @param editable
	 *            whether the components should be enable for editing or not
	 * @param list 
	 */
	public RshellInputViewer(ActivityInputPortDefinitionBean bean,
			boolean editable, List<RShellPortSymanticTypeBean> list) {
		this.bean = bean;
		this.editable = editable;
		this.list = list;
		setBorder(javax.swing.BorderFactory.createEtchedBorder());
		initView();
		setEditMode();
	}

	/**
	 * Uses {@link GridBagLayout} to layout the overall component. Adds the
	 * individual editable elements to the view to allow parts of the
	 * {@link ActivityInputPortDefinitionBean} to be changeD
	 */
	private void initView() {
		setLayout(new GridBagLayout());
		GridBagConstraints outerConstraint = new GridBagConstraints();
		outerConstraint.anchor = GridBagConstraints.FIRST_LINE_START;
		outerConstraint.gridx = 0;
		outerConstraint.gridy = 0;
		outerConstraint.weighty = 0;
		outerConstraint.weightx = 0.1;
		outerConstraint.fill = GridBagConstraints.BOTH;

		nameField = new JTextField(bean.getName());
		add(nameField, outerConstraint);
		
		semanticTypeSelector = new JComboBox(RshellPortTypes.getInputSymanticTypes());
		semanticTypeSelector.setSelectedItem(SemanticTypes.STRING);
		
		semanticTypeSelector.setRenderer(new PortTypesListCellRenderer());
		
		for (RShellPortSymanticTypeBean inputType : list) {
			if (bean.getName().equalsIgnoreCase(inputType.getName())) {
				semanticTypeSelector.setSelectedItem(inputType.getSymanticType());
				break;
			}
		}
		
		outerConstraint.gridx = 1;
		add(semanticTypeSelector, outerConstraint);

		outerConstraint.gridx = 2;
		SpinnerNumberModel model = new SpinnerNumberModel(new Integer(bean
				.getDepth()), new Integer(0), new Integer(100), new Integer(1));
		depthSpinner = new JSpinner(model);
		depthSpinner.setEnabled(false);
		depthSpinner.setToolTipText("A depth of 0 means a simple value, like a string. Depth 1 is a list of simple values, while depth 2 is a list of a list of simple values");
		// depthSpinner.setValue(bean.getDepth());

		add(depthSpinner, outerConstraint);

		outerConstraint.gridx = 3;
		refSchemeText = new DialogTextArea();
//		String refs = "";
////		for (Object refScheme : bean.getHandledReferenceSchemes()) {
////			refs = refs + refScheme.getClass().getSimpleName() + "\n";
////		}
//		refSchemeText.setText(refs);
//		refSchemeText.setEditable(false);
//		refSchemeText.setBorder(javax.swing.BorderFactory.createEtchedBorder());
		// add(refSchemeText, outerConstraint);

		outerConstraint.gridx = 4;
		// mimeTypeText = new DialogTextArea();
		// String mimes = "";
		// for (String mimeType : bean.getMimeTypes()) {
		// mimes = mimes + mimeType + "\n";
		// }
		// mimeTypeText.setText(mimes);
		// mimeTypeText.setEditable(false);
		// mimeTypeText.setBorder(javax.swing.BorderFactory.createEtchedBorder());
		// add(mimeTypeText, outerConstraint);

		outerConstraint.gridx = 5;
//		translatedType = new JLabel(bean.getTranslatedElementType()
//				.getSimpleName());
		translatedType = new JLabel();
		// add(translatedType, outerConstraint);
	}

	/**
	 * Get the component which allows the
	 * {@link ActivityInputPortDefinitionBean} name to be edited
	 * 
	 * @return
	 */
	public JTextField getNameField() {
		return nameField;
	}

	/**
	 * Get the component which allows {@link ReferenceScheme}s to be added to a
	 * {@link ActivityInputPortDefinitionBean}
	 * 
	 * @return
	 */
	public DialogTextArea getRefSchemeText() {
		return refSchemeText;
	}

	/**
	 * Get the component which allows Mime Types to be added to a
	 * {@link ActivityInputPortDefinitionBean}
	 * 
	 * @return
	 */
	public DialogTextArea getMimeTypeText() {
		return mimeTypeText;
	}

	/**
	 * Get the component which allows Translated Types to be added to a
	 * {@link ActivityInputPortDefinitionBean}
	 * 
	 * @return
	 */
	public JLabel getTranslatedType() {
		return translatedType;
	}

	/**
	 * Set the
	 * {@link ActivityInputPortDefinitionBean#setAllowsLiteralValues(boolean)}
	 * to be true or false using this component
	 * 
	 * @return
	 */
	public JComboBox getSemanticSelector() {
		return semanticTypeSelector;
	}

	/**
	 * Change the depth of the {@link ActivityInputPortDefinitionBean}
	 * 
	 * @return
	 */
	public JSpinner getDepthSpinner() {
		return depthSpinner;
	}

	/**
	 * Get the actual {@link ActivityInputPortDefinitionBean} which is
	 * represented by this view
	 * 
	 * @return
	 */
	public ActivityInputPortDefinitionBean getBean() {
		return bean;
	}

	/**
	 * Can the components on this view be edited?
	 * 
	 * @return
	 */
	public boolean isEditable() {
		return editable;
	}

	/**
	 * Set all the components to be editable or not
	 * 
	 * @param editable
	 */
	public void setEditable(boolean editable) {
		this.editable = editable;
		setEditMode();
	}

	/**
	 * Sets the {@link #nameField}, {@link #semanticTypeSelector} and
	 * {@link #depthSpinner} to allow editing
	 */
	public void setEditMode() {
		this.nameField.setEditable(editable);
		this.semanticTypeSelector.setEnabled(editable);
		this.depthSpinner.setEnabled(editable);

	}


}
